@@ -976,14 +976,14 @@ let () =
      macro body [e] *)
   let suspend = function
     | Macro.Uid m -> Suspend.add m
-    | Macro.Lid _ -> () in
+    | Macro.Lid _ | Macro.Macro _ | Macro.Unused -> () in
   Macro.Hook.define_params (fun params -> List.iter suspend params);
   (* After the macro body is read, remove the suspended modules.  If
      an argument is repeated, the overloading will be removed the
      same number of times (this is important). *)
   let unsuspend = function
     | Macro.Uid m -> Suspend.remove m;
-    | Macro.Lid _ -> () in
+    | Macro.Lid _  | Macro.Macro _ | Macro.Unused -> () in
   Macro.Hook.define (fun params body -> List.iter unsuspend params; body);
   Macro.Hook.macro_expansion Suspend.resolve
 ;;
