@@ -122,7 +122,7 @@
 
 
     @author Dany Maslowski, Julie De Pril, Christophe Troestler
-    @version 0.8.8
+    @version 0.8.10
 *)
 
 open Camlp4.PreCast.Syntax.Ast
@@ -305,7 +305,9 @@ val lid_subst : t -> (string * string) list -> t
 
 val std_arithmetic : t -> t
   (** [std_arithmetic t] is just a shortcut for [lid_subst t
-      [("+","add"); ("-","sub"); ("~-", "neg"); ("*","mul"); ("/","div")]] *)
+      [("+","add"); ("-","sub"); ("~-", "neg"); ("*","mul");
+      ("/","div")]].  Note that the unary negation, [~-], is
+      overloaded as well. *)
 
 val comparison : ?cache: bool -> ?cmp:string -> t -> t
   (** [comparison ?cmp t] return the set of overloadings [t] augmented
@@ -406,7 +408,9 @@ val apply : t -> module_longident -> (expr -> expr)
 
 val associate : t -> ?qualify:string -> string -> unit
   (** [associate_t t m] links the "set of overloadings" [t] to the
-      module longident [m].
+      module longident [m].  For example [let t = std_arithmetic empty
+      in associate t "I"] implies that [I.(x + y * z)] will be
+      transformed into [I.add x (I.mul y z)].
 
       @param qualify allows to specify a module different from [m] for
       the qualidifcation of the relocatable expressions in the
