@@ -162,9 +162,7 @@ let rec module_longident_of_id name loc id =
  *                              Hooks
  ***********************************************************************)
 
-type param = Lid of string | Uid of string
-  (* Types of arguments.  [Lid] are supposed to be substituted by
-     expressions while [Uid] are for module replacement. *)
+type param = Lid of string | Uid of string | Macro of string | Unused
 
 (* Other macro packages may need some way of cooperating with macros,
    especially to delay their expansion with IFDEF.  This is why we
@@ -194,27 +192,77 @@ struct
 (* FIXME: More hooks needed. *)
 end
 
+(***********************************************************************
+ *                          Macro storage
+ ***********************************************************************)
+
+exception Invalid of string
+
+(* Control which macros can be expanded.  In the body of a DEFINE, the
+   parameters should not be expanded even if they are macros.  In a
+   IFDEF branch, no macro is expanded because the branch can define
+   other macros and we do not want that substituting these macros
+   affects the body of already substituted macros.  Example:
+   [DEFINE X = A  IFDEF X THEN  DEFINE A = 1  DEFINE B = X END]. *)
+module Allow =
+struct
+  let suspend_level = ref 0
+    (* to be able to suspend macro replacement (see [eval]). *)
+  let not_for = Hashtbl.create 10
+
+  let subst m = !suspend_level = 0 && not(Hashtbl.mem not_for m)
+
+  (* We use a counter because there may be several groups. *)
+  let disable () = incr suspend_level
+  let reenable () = decr suspend_level
+
+  (* No macros embedded in other => no need to count groupings *)
+  let remove m = Hashtbl.add not_for m 0
+  let clear () = Hashtbl.clear not_for
+end
+
+(* Defined macros (global value).  *)
+let defined = Hashtbl.create 10
+
+let is_defined_macro name = Hashtbl.mem defined name
+
+let is_defined name =
+  if is_uppercase_identifier name then is_defined_macro name
+  else raise(Invalid name)
 
 (***********************************************************************
  *                           Substitution
  ***********************************************************************)
 
-(* The format is compatible with Emacs compilation mode.  *)
-let fname_with_loc loc msg =
+(* As [loc_to_string] but without the "File" to also use it in
+   [fname_with_loc].  The format is compatible with Emacs compilation mode. *)
+let partial_loc_to_string loc =
   let (fname, line0, bol0, off0, line1, bol1, off1, _) = Loc.to_tuple loc in
   let line =
     if line0 = line1 then sprintf "%i" line0
     else sprintf "%i-%i" line0 line1 in
-  (* "File \"" added by Loc.to_string *)
-  sprintf "%s\", line %s, characters %i-%i:\n  %s\n  File \"%s"
-    fname line (off0 - bol0) (off1 - bol1) msg fname
+  (* "File \"" added by [Loc.to_string] or [loc_to_string] *)
+  sprintf "%s\", line %s, characters %i-%i:"
+    fname line (off0 - bol0) (off1 - bol1)
 
-class reloc_fname loc msg =
-  let fname_loc = fname_with_loc loc msg in
+(* As of OCaml 3.11.0, Loc.to_string returns location strings
+   incompatible with Emacs 22.3.  Thus we use: *)
+let loc_to_string loc =
+  "File \"" ^ (partial_loc_to_string loc)
+
+(** [(new reloc_fname use_loc msg)#expr e] prefix all locations in the
+    expression [e] with the llocation [use_loc] (typically the
+    location where the macro is used) and the message [msg]. *)
+class reloc_fname use_loc msg =
+  let fname = partial_loc_to_string use_loc in
+  let fname_with_loc _loc =
+    let fname_loc =
+      sprintf "%s\n  %s\n  File \"%s" fname msg (Loc.file_name _loc) in
+    Loc.set_file_name fname_loc _loc in
 object
   inherit Ast.map as super
-  method loc _loc = Loc.set_file_name fname_loc _loc
-  method _Loc_t _loc = Loc.set_file_name fname_loc _loc (* camlp4 3.10.0 *)
+  method loc _loc = fname_with_loc _loc
+  method _Loc_t _loc = fname_with_loc _loc (* camlp4 3.10.0 *)
 end
 
 (* The original macros were really hard to debug because the location
@@ -223,38 +271,119 @@ end
    while the original locations are preserved so that your editor can
    easily send you to the point where the problem occurs inside the
    body of the macro. *)
-class reloc macro use_loc =
+class reloc use_loc macro =
   let fname_use = sprintf "Expanding of the macro \
       %S at the previous location yields the error:" macro in
   reloc_fname use_loc fname_use
 
-class subst_macro macro use_loc env_var env_mod =
+type param_values = {
+  lid : (string * Ast.expr) list; (* replacements for lid *)
+  uid : (string * string list) list; (* replacements for modules params *)
+  macro : (string * (Loc.t * string)) list; (* replacements for macros *)
+}
+
+(* Split and check the arguments depending on whether the variables
+   are lowercase or uppercase. *)
+let split_args use_loc macro args values macro_loc =
+  let rec loop env a v =
+    match a, v with
+    | [], [] -> env
+    | (Lid a) :: atl, v :: vtl ->
+        loop { env with lid = (a, v) :: env.lid } atl vtl
+    | (Uid a) :: atl, v :: vtl ->
+        let i = module_longident_of_id a use_loc v in
+        loop { env with uid = (a, i) :: env.uid } atl vtl
+    | Macro a :: atl, v :: vtl ->
+        let uid = match v with
+          | <:expr@loc< $uid:s$ >> when is_uppercase_identifier s -> (loc, s)
+          | _ ->
+              let msg = sprintf "can only substitute the parameter \"%s()\" of \
+                the macro %S\n  with an uppercase identifier. See\n  \
+                %s definition of %S."
+                a macro (loc_to_string macro_loc) macro in
+              Loc.raise use_loc (Stream.Error msg)
+        in
+        loop { env with macro = (a, uid) :: env.macro } atl vtl
+    | Unused :: atl, _ :: vtl ->
+        loop env atl vtl (* ignore the param value *)
+    | _ ->
+        let msg = sprintf "expected %d parameters; found %d.  \
+          See\n  %s definition of %S."
+          (List.length args) (List.length values)
+          (loc_to_string macro_loc) macro in
+        Loc.raise use_loc (Stream.Error msg)
+  in
+  loop { lid = []; uid = []; macro = [] } args values
+
+exception Not_constructor
+
+(** [get_constructor e] returns [(uid, args)] if [e] is (syntactically) a
+    constructor.  @raise Not_constructor if [e] is not a constructor. *)
+let get_constructor =
+  (* The constructors are represented in curried form in Camlp4. *)
+  let rec get = function
+  | <:expr< $uid:u$ >> -> (u, [])
+  | <:expr< $c$ $a$ >> -> let (u, args) = get c in (u, a :: args)
+  | _ -> raise Not_constructor
+  in
+  fun e -> let (u, args) = get e in (u, List.rev args)
+
+(** [subst_macro use_loc macro params values body] transversal to
+    substitute the parameters [params] with their corresponding
+    [values] in the [body] of the macro. *)
+class subst_macro use_loc macro params values body =
+  let env = split_args use_loc macro params values (Ast.loc_of_expr body) in
 object(self)
-  inherit reloc macro use_loc as super
+  inherit reloc use_loc macro as super
+
+  method perform =
+    (* First substitute all macro parameters with their values
+       [self#expr body], only then apply the (now well qualified)
+       overloadings or other hooks. *)
+    Hook.apply_macro_expansion (self#expr body)
 
   method expr = function
   | <:expr@_loc< $lid:x$ >> as e ->
     (try
-       let e' = List.assoc x env_var in
+       let e' = List.assoc x env.lid in
        (* We substitute [x] by the expression [e'].  We want to
           preserve the locations of [e'] in case the error is inside [e']. *)
        let arg_loc = new reloc_fname _loc (
-         sprintf "Expanding the macro %S, substituting %S at the previous \
-           location:" macro x) in
+         sprintf "Expanding the macro %S, substituting the variable %S \
+           yields:" macro x) in
        arg_loc#expr e'
      with Not_found -> super#expr e)
-  | e -> super#expr e
+  | e ->
+    (try
+       let macro, values = get_constructor e (* or Not_constructor *) in
+       let macro_loc, macro = List.assoc macro env.macro (* or Not_found *) in
+       (* Transform params values first: *)
+       let values = List.map self#expr values in
+       (try
+          (* Get the macro from the GLOBAL environment *)
+          match Hashtbl.find defined macro with
+          | Some([], _) | None ->
+              let msg = sprintf "The macro %S must accept parameters." macro in
+              Loc.raise macro_loc (Stream.Error msg)
+          | Some(params, body) ->
+              let loc = Ast.loc_of_expr e in
+              (new subst_macro loc macro params values body)#perform
+        with Not_found ->
+          let msg = sprintf "The macro %S must be previously defined." macro in
+          Loc.raise macro_loc (Stream.Error msg))
+     with Not_found | Not_constructor -> super#expr e)
 
   method ident i = match i with
   | <:ident@loc< $uid:m$ >> ->
     (try
-       let m' = List.assoc m env_mod in
+       let m' = List.assoc m env.uid in
        Module_longident.to_id m' (self#loc loc)
      with Not_found -> i)
   | _ -> super#ident i
 end
 
 
+
 let bad_patt _loc =
   let msg = "This macro cannot be used in a pattern (see its definition)" in
   Loc.raise _loc (Failure msg)
@@ -288,84 +417,16 @@ let patt_of_expr env =
  *                         Macro definition
  ***********************************************************************)
 
-exception Invalid of string
-
-(* Control which macros can be expanded.  In the body of a DEFINE, the
-   parameters should not be expanded even if they are macros.  In a
-   IFDEF branch, no macro is expanded because the branch can define
-   other macros and we do not want that substituting these macros
-   affects the body of already substituted macros.  Example:
-   [DEFINE X = A  IFDEF X THEN  DEFINE A = 1  DEFINE B = X END]. *)
-module Allow =
+module Macro_params =
 struct
-  let suspend_level = ref 0
-    (* to be able to suspend macro replacement (see [eval]). *)
-  let not_for = Hashtbl.create 10
-
-  let subst m = !suspend_level = 0 && not(Hashtbl.mem not_for m)
-
-  (* We use a counter because there may be several groups. *)
-  let disable () = incr suspend_level
-  let reenable () = decr suspend_level
-
-  (* No macros embedded in other => no need to count groupings *)
-  let remove m = Hashtbl.add not_for m 0
-  let clear () = Hashtbl.clear not_for
+  let level_inside = ref 0
+    (* global variable that tells whether the current parsing position is
+       inside macro parameters values. *)
+  let enter () = incr level_inside
+  let leave () = decr level_inside
+  let is_outside () = !level_inside = 0
 end
 
-(* Split ang check the arguments depending on whether the variables
-   are lowercase or uppercase. *)
-let split_args loc args values =
-  let rec loop env_var env_mod a v =
-    match a, v with
-    | [], [] -> env_var, env_mod
-    | (Lid a) :: atl, v :: vtl ->
-        loop ((a, v) :: env_var) env_mod atl vtl
-    | (Uid a) :: atl, v :: vtl ->
-        let i = module_longident_of_id a loc v in
-        loop env_var ((a, i) :: env_mod) atl vtl
-    | _ ->
-        let msg = sprintf "expected %d parameters; found %d"
-          (List.length args) (List.length values) in
-        Loc.raise loc (Failure msg)
-  in
-  loop [] [] args values
-
-
-(* Defined macros.  *)
-let defined = Hashtbl.create 10
-
-let is_defined_macro name = Hashtbl.mem defined name
-
-let is_defined name =
-  if is_uppercase_identifier name then is_defined_macro name
-  else raise(Invalid name)
-
-let is_defined_params name =
-  try
-    (match Hashtbl.find defined name with
-     | Some(_ :: _, _) -> true
-     | _ -> false)
-  with Not_found -> false
-
-let is_defined_0param name =
-  try
-    (match Hashtbl.find defined name with
-     | Some([], _) -> true
-     | _ -> false)
-  with Not_found -> false
-
-
-let subst_macro_with_params _loc name args params body =
-  let params = match params with
-    | <:expr< ($tup:e$) >> -> Ast.list_of_expr e []
-    | e -> [e]  in
-  let env_var, env_mod = split_args _loc args params in
-  (* First substitute all macro parameters, only then apply the (now
-     well qualified) overloadings or other hooks. *)
-  let e = (new subst_macro name _loc env_var env_mod)#expr body in
-  Hook.apply_macro_expansion e
-
 let define_macro name args_expr =
   begin match args_expr with
   | Some([], e) ->
@@ -373,26 +434,43 @@ let define_macro name args_expr =
         GLOBAL: Syntax.expr Syntax.patt;
       Syntax.expr: LEVEL "simple"
         [ [ UIDENT $name$ ->
-              if Allow.subst name then (new reloc name _loc)#expr e
+              if Allow.subst name then (new reloc _loc name)#expr e
               else <:expr< $uid:name$ >> ] ];
       Syntax.patt: LEVEL "simple"
         [ [ UIDENT $name$ ->
               if Allow.subst name then
                 (* FIXME: does not work *)
                 let p = patt_of_expr [] e in
-                (new reloc name _loc)#patt p
+                (new reloc _loc name)#patt p
               else <:patt< $uid:name$ >>       ] ];
       END
-  | Some(args, e) ->
+  | Some(params, body) ->
       EXTEND Gram
-        Syntax.expr: LEVEL "apply"
-        [ [ UIDENT $name$; params = SELF ->
+        GLOBAL: Syntax.expr Syntax.patt;
+
+      values_enter: [ [ -> Macro_params.enter() ] ];
+      Syntax.expr: LEVEL "apply"
+        (* Force the parameter values of the macro to be inside braces
+           (this also allows the second case not to be a particular
+           case of the first): *)
+        [ [ UIDENT $name$; "("; _ = values_enter; values = SELF; ")" ->
+              Macro_params.leave();
               if Allow.subst name then
                 (* The parameters have been read while macro expansion
                    was enabled => no need to transform them. *)
-                subst_macro_with_params _loc name args params e
+                let values = match values with
+                  | <:expr< $tup:e$ >> -> Ast.list_of_expr e []
+                  | e -> [e] in
+                (new subst_macro _loc name params values body)#perform
               else
-                <:expr< $uid:name$ $params$ >>
+                <:expr< $uid:name$ $values$ >>
+          | UIDENT $name$ ->
+              if Macro_params.is_outside() then (
+                let msg = sprintf "Warning: The macro %S exists but requires \
+                arguments, left unchanged." name in
+                print_warning _loc msg;
+              );
+              <:expr< $uid:name$ >>
           ] ];
       (*Syntax.patt: LEVEL "simple"
         [ [ UIDENT $name$; param = SELF ->
@@ -409,17 +487,22 @@ let define_macro name args_expr =
   end;
   Hashtbl.add defined name args_expr
 
+let bad_param test s =
+  if not test then
+    invalid_arg(sprintf "Pa_do.Macro.define: bad param name %S." s)
+
 let define ?expr name =
   if is_uppercase_identifier name then
     let expr = match expr with
       | None -> None
       | Some(args, e) ->
-          let classify s =
-            if is_lowercase_identifier s then Lid s
-            else if is_capitalized_identifier s then Uid s
-            else invalid_arg(sprintf "Pa_do.Macro.define: bad param name %S."
-                               s) in
-          Some(List.map classify args, e)
+          let check = function
+            | Lid s -> bad_param (is_lowercase_identifier s) s
+            | Uid s -> bad_param (is_capitalized_identifier s) s
+            | Macro s -> bad_param (is_uppercase_identifier s) s
+            | Unused -> () in
+          List.iter check args;
+          Some(args, e)
     in
     define_macro name expr
   else raise(Invalid name)
@@ -454,19 +537,39 @@ let add_include_dir dir =
   if dir <> "" then
     include_dirs := dir :: !include_dirs
 
-let parse_include_file =
-  let exists file dir = Sys.file_exists (Filename.concat dir file) in
-  fun entry file ->
-    let file =
-      try
-        let dir = List.find (exists file) (List.rev !include_dirs) in
-        Filename.concat dir file
-      with Not_found -> file (* absolute path or relative to current dir *) in
-    let ch = open_in file in
-    let st = Stream.of_channel ch in
-    let g = Gram.parse entry (Loc.mk file) st in
-    close_in ch;
-    g
+let file_exists file dir = Sys.file_exists (Filename.concat dir file)
+
+let parse_include_file entry file =
+  let file =
+    try
+      let dir = List.find (file_exists file) (List.rev !include_dirs) in
+      Filename.concat dir file
+    with Not_found -> file (* absolute path or relative to current dir *) in
+  let ch = open_in file in
+  let st = Stream.of_channel ch in
+  let g = Gram.parse entry (Loc.mk file) st in
+  close_in ch;
+  g
+
+let rec buffer_of_channel content buf len ch =
+  let read = input ch buf 0 len in
+  if read = 0 then content (* End of file *)
+  else (
+    Buffer.add_substring content buf 0 read;
+    buffer_of_channel content buf len ch
+  )
+
+let include_as_string file =
+  let file =
+    try
+      let dir = List.find (file_exists file) (List.rev !include_dirs) in
+      Filename.concat dir file
+    with Not_found -> file (* absolute path or relative to current dir *) in
+  let ch = open_in file in
+  let buf = String.create 0x1000 in
+  let content = buffer_of_channel (Buffer.create 0x1000) buf 0x1000 ch in
+  close_in ch;
+  Buffer.contents content
 
 
 (***********************************************************************
@@ -492,29 +595,38 @@ type 'a macro_or_item =
    (this behavior is different from the macros provided with camlp4
    but is considered better).  *)
 
-(* [expand_macros] is a class to expand currently defined macros in
-   expressions. *)
+(** [expand_macros] is a class to expand currently defined macros (in
+    the global environment) in expressions. *)
 let expand_macros =
 object(self)
   inherit Ast.map as super
 
   (* See [define_macro] *)
-  method expr = function
-  | <:expr@_loc< $uid:name$ $params$ >>
-      when is_defined_params name && Allow.subst name ->
-    (match Hashtbl.find defined name with
-     | Some(args, body) ->
-         (* The [params] have been read _without_ macro expansion =>
-            first recurse to transform them *)
-         let params = self#expr params in
-         subst_macro_with_params _loc name args params body
-     | _ -> assert false)
-  | <:expr@_loc< $uid:name$ >>
-      when is_defined_0param name && Allow.subst name ->
-    (match Hashtbl.find defined name with
-     | Some([], body) -> (new reloc name _loc)#expr body
-     | _ -> assert false)
-  | e -> super#expr e
+  method expr e =
+    let _loc = Ast.loc_of_expr e in
+    try
+      let macro, values = get_constructor e (* or Not_constructor *) in
+      if is_defined_macro macro && Allow.subst macro then
+        (* The [values] have been read _without_ macro expansion =>
+           first recurse to transform them. *)
+        let values = List.map self#expr values in
+        match Hashtbl.find defined macro with
+        | Some([], body) ->
+            (match (new reloc _loc macro)#expr body with
+             | <:expr< $uid:_$ >> ->
+               (* constructor => currying values (for Campl4) *)
+               List.fold_left (fun c a -> <:expr< $c$ $a$ >>) body values
+             | _ ->
+                 (* Not constructor => tuple (only if more than one param) *)
+                 match values with
+                 | [] -> body
+                 | [p] -> <:expr< $body$ $p$ >>
+                 | _ -> <:expr< $body$ $tup: Ast.exCom_of_list values$ >>)
+        | Some(params, body) ->
+            (new subst_macro _loc macro params values body)#perform
+        | None -> super#expr e
+      else super#expr e
+    with Not_constructor -> super#expr e
 
 (* FIXME: patt *)
 end
@@ -576,6 +688,7 @@ open Syntax
 let check_duplicate_params =
   let rec has_dup = function
     | [] | [_] -> false
+    | Unused :: tl -> has_dup tl
     | a :: ((b :: _) as tl) -> a = b || has_dup tl in
   fun _loc pl ->
     if has_dup (List.sort compare pl) then (
@@ -595,8 +708,8 @@ let () =
   opt_macro_value:
     [ [ "("; pl = macro_params; ")"; "="; e = expr ->
           (* Allow again the expansion of macros with names identical
-             to parameters.  No macro can be embedded into another =>
-             no multiple levels to take care. *)
+             to parameters.  No macro definition can be embedded into
+             another => no multiple levels to take care. *)
           (Allow.clear();
            Some (pl, Hook.apply_define pl e))
       | "="; e = expr -> Some ([], Hook.apply_define [] e)
@@ -610,11 +723,15 @@ let () =
       ] ];
   macro_1param:
     [ [ x = LIDENT -> Lid x
-      | x = UIDENT ->
+      | x = UIDENT; "("; ")" ->
           (* Do not allow [x] to be expanded in the macro body even if
              it is a macro name. *)
           Allow.remove x;
-          Uid x  ] ];
+          Macro x
+      | x = UIDENT ->
+          Allow.remove x;
+          Uid x
+      | "_" -> Unused ] ];
   macro_statement:
     [ [ "DEFINE"; i = UIDENT; def = opt_macro_value -> Def(i, def)
       | "UNDEF"; i = UIDENT -> Undef i
@@ -648,6 +765,8 @@ let () =
           if is_defined_macro i then e1 else e2
       | "IFNDEF"; i = UIDENT; "THEN"; e1 = expr; e2 = else_expr ->
           if is_defined_macro i then e2 else e1
+      | "INCLUDE_AS_STRING"; fname = STRING ->
+          <:expr< $str:include_as_string fname$ >>
       ] ];
   else_expr:
     [ [ "ELSE"; e = expr; endif -> e
